;; Generated package description from visual-filename-abbrev.el  -*- no-byte-compile: t -*-
(define-package "visual-filename-abbrev" "1.3" "Visually abbreviate filenames" '((emacs "26.1")) :commit "4af7868db7bbc6a3c185540ba4e257d4c4c560c0" :url "https://elpa.gnu.org/packages/visual-filename-abbrev.html" :authors '(("Tassilo Horn" . "tsdh@gnu.org")) :maintainer '("Tassilo Horn" . "tsdh@gnu.org") :keywords '("tools"))
